/*
 * This file is part of PAXLink, which is an AI for OpenTTD
 * Copyright (C) 2008, 2009  Leif Linse
 *
 * PAXLink is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License
 *
 * PAXLink is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with PAXLink; If not, see <http://www.gnu.org/licenses/> or
 * write to the Free Software Foundation, Inc., 51 Franklin Street, 
 * Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
require("helper.nut");
require("globals.nut");
require("roadbuilder.nut");
require("orders.nut");


//////////////////////////////////////////////////////
//                                                  //
//   CLASS: Builder                                 //
//                                                  //
//////////////////////////////////////////////////////
class Builder {
	conf_max_buy_num_buses_per_bus_station = null;
	conf_expand_if_more_than_this_num_buses_per_bus_station = null;

	constructor()
	{
		conf_max_buy_num_buses_per_bus_station = 4.5; 
		conf_expand_if_more_than_this_num_buses_per_bus_station = 4.0;
	}
}

function Builder::MyValuate(list, valuator, ...)
{
   assert(typeof(list) == "instance");
   assert(typeof(valuator) == "function");
   
   local args = [this, null];
   
   for(local c = 0; c < vargc; c++) {
      args.append(vargv[c]);
   }

   foreach(item, _ in list) {
      args[1] = item;
      local value = valuator.acall(args);
      if (typeof(value) == "bool") {
         value = value ? 1 : 0;
      } else if (typeof(value) != "integer") {
         throw("Invalid return type from valuator");
      }
      list.SetValue(item, value);
   }
}

function GetBuildGridSize(town_id)
{
	local pop = AITown.GetPopulation(town_id);

	local size = 1 + pop / 2000;
	size = size.tointeger();

	//AILog.Info("Suggests grid size " + size);

	return size;
}

function MakeTownGrid(town_id, size)
{
	local town_center = AITown.GetLocation(town_id);

	local town_center_x = AIMap.GetTileX(town_center);
	local town_center_y = AIMap.GetTileY(town_center);

	local ret = AIList();

	local x, y;
	for(x = town_center_x - 7 * size + 3; x <= town_center_x + 7 * size - 3; x += 7)
	{
		for(y = town_center_y - 7 * size + 3; y <= town_center_y + 7 * size - 3; y += 7)
		{
			local tile = AIMap.GetTileIndex(x, y);

			// only add tile if it is closest to town_id town
			if(AITile.GetClosestTown(tile) == town_id)
			{
				ret.AddItem(tile, 0);
			}
		}
	}

	return ret;
}

function IsTileOnTownRoadGrid(tile_id)
{
	local n = AIGameSettings.GetValue("economy.town_layout").tointeger();
	AILog.Error("town_layout: " + n);
	local dx = abs(tile_id % AIMap.GetMapSizeY());
	local dy = abs(tile_id / AIMap.GetMapSizeY());
	local on_road_grid = (n == 2 || n == 3) &&     // Town has 2x2 or 3x3 road grid layout
		(dx % (n+1) == 0 || dy % (n+1) == 0);      // and tile is on the road grid

	return on_road_grid;
}

function GetTileRelative(relative_to_tile, delta_x, delta_y)
{
	local tile_x = AIMap.GetTileX(relative_to_tile);
	local tile_y = AIMap.GetTileY(relative_to_tile);

	local new_x = tile_x + delta_x;
	local new_y = tile_y + delta_y;

	return AIMap.GetTileIndex(new_x, new_y);
}

function GetNeighbours4(tile_id)
{
	local list = AITileList();

	if(!AIMap.IsValidTile(tile_id))
		return list;

	local tile_x = AIMap.GetTileX(tile_id);
	local tile_y = AIMap.GetTileY(tile_id);

	list.AddTile(GetTileRelative(tile_id, -1, 0));
	list.AddTile(GetTileRelative(tile_id, 0, -1));
	list.AddTile(GetTileRelative(tile_id, 1, 0));
	list.AddTile(GetTileRelative(tile_id, 0, 1));

	return list;
}

function GetNeighbours8(tile_id)
{
	local list = AITileList();

	if(!AIMap.IsValidTile(tile_id))
		return list;

	local tile_x = AIMap.GetTileX(tile_id);
	local tile_y = AIMap.GetTileY(tile_id);

	list.AddTile(GetTileRelative(tile_id, -1, 0));
	list.AddTile(GetTileRelative(tile_id, 0, -1));
	list.AddTile(GetTileRelative(tile_id, 1, 0));
	list.AddTile(GetTileRelative(tile_id, 0, 1));

	list.AddTile(GetTileRelative(tile_id, -1, -1));
	list.AddTile(GetTileRelative(tile_id, 1, -1));
	list.AddTile(GetTileRelative(tile_id, 1, 1));
	list.AddTile(GetTileRelative(tile_id, -1, 1));

	return list;
}

function GridTileNeighbourScoreValuator(neighbour_tile_id, grid_tile_id)
{
	local score = 0;

	if(AITile.IsBuildable(neighbour_tile_id))
		score += 100;

	if(AIRoad.GetNeighbourRoadCount(neighbour_tile_id) != 0)
		score += 200;

	local diff_x = AIMap.GetTileX(grid_tile_id) - AIMap.GetTileX(neighbour_tile_id);
	local diff_y = AIMap.GetTileY(grid_tile_id) - AIMap.GetTileY(neighbour_tile_id);
	local opposite_tile_id = grid_tile_id + AIMap.GetTileIndex(diff_x, diff_y);

	// don't give bonus to end of road.
	if(AIRoad.IsRoadTile(opposite_tile_id) && AIRoad.AreRoadTilesConnected(opposite_tile_id, grid_tile_id))
		score -= 200;

	// prefer tiles close to grid_tile_id
	score -= AIMap.DistanceManhattan(grid_tile_id, neighbour_tile_id) * 10

	// Give a big penalty to tiles that is on the town grid
	if(AIController.GetSetting("avoid_town_road_grid") == 1 && IsTileOnTownRoadGrid(neighbour_tile_id))
		score -= 400;

	return score;
}

function CanConnectToRoad(road_tile, adjacent_tile_to_connect)
{
	local neighbours = GetNeighbours4(road_tile);
	
	neighbours.RemoveTile(adjacent_tile_to_connect);
	local fail = false;

	// This function requires that road_tile is connected with at least one other road tile.
	neighbours.Valuate(AIRoad.IsRoadTile);
	if(Helper.ListValueSum(neighbours) < 1)
		return false;
	
	for(local neighbour_tile_id = neighbours.Begin(); neighbours.HasNext(); neighbour_tile_id = neighbours.Next())
	{
		if(AIRoad.IsRoadTile(neighbour_tile_id))
		{
			fail = fail || !AIRoad.CanBuildConnectedRoadPartsHere(road_tile, neighbour_tile_id, adjacent_tile_to_connect);
		}
	}

	return !fail;
}

function DoesSlopeAllowBuildingStationDepotHere(tile_id, front)
{
	local slope = AIMap.GetSlope(tile_id);

	if(slope != 0 && AIMap.IsSteepSlope(tile_id))
		return false;

	local diff_x = AIMap.GetTileX(tile_id) - AIMap.GetTileX(front);
	local diff_y = AIMap.GetTileY(tile_id) - AIMap.GetTileY(front);

	if(diff_x == 1 && diff_y == 0)
	{
	}

	return true;
}

function Builder::BuildRoadStation(tile_id, depot)
{
	//AISign.BuildSign(tile_id, "Bld stn");
	AIRoad.SetCurrentRoadType(AIRoad.ROADTYPE_ROAD);

	if(AITile.IsWaterTile(tile_id))
		return false;

	// Don't even try if the tile has a steep slope
	local slope = AITile.GetSlope(tile_id);
	if(slope != 0 && AITile.IsSteepSlope(slope))
		return false;

	if(AIController.GetSetting("forbid_town_road_grid") == 1 && IsTileOnTownRoadGrid(tile_id))
		return false;

	local neighbours = GetNeighbours4(tile_id);
	neighbours.Valuate(AIRoad.IsRoadTile);
	neighbours.KeepValue(1);

	MyValuate(neighbours, CanConnectToRoad, tile_id);
	neighbours.KeepValue(1);

	//MyValuate(neighbours, AITile.IsSteepSlope); // steep sloped neighbours can't be used
	//neighbours.KeepValue(0);

	neighbours.Valuate(Helper.GetMaxHeightOfTile);
	neighbours.KeepValue(Helper.GetMaxHeightOfTile(tile_id));

	// Return false if none of the neighbours has road
	if(neighbours.Count() == 0)
		return false;

	local front = neighbours.Begin();

	// Hack for now, don't allow any slope until we can detect which slopes are okay.
	//if(AITile.GetSlope(tile_id) != 0)
	//	return false;


	
	/*{ {
		local test = AITestMode();
		local fail = false;


		if(depot)
			fail = !AIRoad.BuildRoadDepot(tile_id, front);
		else
			fail = !AIRoad.BuildRoadStation(tile_id, front, truck, join_adjacent);
		if(fail)
		{
			return false;
		}
	} }*/
	
	if(!CanConnectToRoad(front, tile_id))
		return false;

	//if(!DoesSlopeAllowBuildingStationDepotHere(tile_id, front))
	//	return false;
		
	if(!AITile.IsBuildable(tile_id) && !AIRoad.IsRoadTile(tile_id) && !AICompany.IsMine(AITile.GetOwner(tile_id)))
	{
		AITile.DemolishTile(tile_id);
	}

	if(!AIRoad.BuildRoad(tile_id, front) && AIRoad.AreRoadTilesConnected(tile_id, front))
		return false;

	Helper.SetSign(tile_id, "tile");
	Helper.SetSign(front, "front");

	if(depot)
		return AIRoad.BuildRoadDepot(tile_id, front);
	else
		return AIRoad.BuildRoadStation(tile_id, front, AIRoad.ROADVEHTYPE_BUS, AIStation.STATION_JOIN_ADJACENT);

}

function Builder::BuildInCityStationCloseTo(tile_id, town_id)
{
	local depot = false

	//if(AITile.IsBuildable(tile_id))
	if(BuildRoadStation(tile_id, depot))
	{
		return tile_id;
	}
	else
	{
		AILog.Info("BulidRoadStation error: " + AIError.GetLastErrorString());
		if(AIError.GetLastError() == AIError.ERR_LOCAL_AUTHORITY_REFUSES)
			return AIMap.GetTileIndex(-1, -1);

		local neighbours = GetNeighbours8(tile_id);

		// only accept neighbours that are in the town
		neighbours.Valuate(AITile.GetClosestTown);
		neighbours.KeepValue(town_id);

		//MyValuate(neighbours, GridTileNeighbourScoreValuator, tile_id);
		//Helper.MyClassValuate(neighbours, this.GridTileNeighbourScoreValuator, this, tile_id);
		neighbours.Valuate(GridTileNeighbourScoreValuator, tile_id);
		neighbours.Sort(AIAbstractList.SORT_BY_VALUE, false);

		if(neighbours.Count() > 0)
		{
			// Try with first neighbour, but if that fails, take next one and try.
			local build_at = neighbours.Begin();
			while(neighbours.HasNext())
			{
				if(BuildRoadStation(build_at, depot))
				{
					AILog.Info("Built road station.");
					return build_at;
				}
				else if(AIError.GetLastError() == AIError.ERR_LOCAL_AUTHORITY_REFUSES)
				{
					// Don't retry if the reason for failure was that the local authority refused building
					return AIMap.GetTileIndex(-1, -1);
				}

				build_at = neighbours.Next();
			}
		}
	}

	//AILog.Info("Returning tile -1, -1");
	return AIMap.GetTileIndex(-1, -1);
}

//// BEGIN OF INTER-CITY helpers

function IsTileRectBuildableAndFlat(top_left_tile, width, height)
{
	local tiles = AITileList();

	// First use API function to check rect as that is faster than doing it self in squirrel.
	if(!AITile.IsBuildableRectangle(top_left_tile, width, height))
		return false;

	// Then only if it is buildable, check that it is flat also.
	tiles.AddRectangle(top_left_tile, GetTileRelative(top_left_tile, width - 1, height - 1));

	// remember how many tiles there are from the beginning
	local count_before = tiles.Count();

	// remove non-flat tiles
	tiles.Valuate(AITile.GetSlope);
	tiles.KeepValue(0);

	// if all tiles are remaining, then all tiles are flat
	return count_before == tiles.Count();
}

//function WillAirportAccept

function FindInterCityStationPlacement(town_id, station_x, station_y)
{
	local town_tiles = AITileList();
	local town_tile = AITown.GetLocation(town_id);
	local radius = 20 + Helper.Max(0, AITown.GetPopulation(town_id) / 1000);
	town_tiles.AddRectangle(GetTileRelative(town_tile, -radius, -radius), GetTileRelative(town_tile, radius, radius));

	/*for(local i = town_tiles.Begin(); town_tiles.HasNext(); i = town_tiles.Next())
	{
		Helper.SetSign(i, " ");
	}*/

	town_tiles.Valuate(AITile.GetClosestTown);
	town_tiles.KeepValue(town_id);

	/*for(local i = town_tiles.Begin(); town_tiles.HasNext(); i = town_tiles.Next())
	{
		//Helper.SetSign(i, "t");
		AISign.BuildSign(i, "t");
	}*/

	town_tiles.Valuate(IsTileRectBuildableAndFlat, station_x, station_y);
	town_tiles.KeepValue(1);

	/*for(local i = town_tiles.Begin(); town_tiles.HasNext(); i = town_tiles.Next())
	{
		//Helper.SetSign(i, "bf");
		AISign.BuildSign(i, "bf");
	}*/


	//Helper.SetSign(town_tiles.Begin(), "here");

	if(town_tiles.IsEmpty())
		return -1;

	town_tiles.Valuate(AIMap.DistanceManhattan, town_tile);
	town_tiles.Sort(AIAbstractList.SORT_BY_VALUE, true); // smallest first

	return town_tiles.Begin();
}

function GetBestAirportAvailable(max_x, max_y)
{
	local airports = [ AIAirport.AT_INTERCON, AIAirport.AT_INTERNATIONAL, AIAirport.AT_METROPOLITAN, AIAirport.AT_LARGE, AIAirport.AT_COMMUTER, AIAirport.AT_SMALL ]

	foreach(airport in airports)
	{
		if(AIAirport.IsValidAirportType(airport) && AIAirport.GetAirportWidth(airport) <= max_x && AIAirport.GetAirportHeight(airport) <= max_y)
			return airport;
	}

	return AIAirport.AT_INVALID;
}

function Builder::GetAllInterCityStations()
{
	local station_list = AIStationList(AIStation.STATION_AIRPORT);
	return station_list;
}

function Builder::FindInterCityStation(town_id)
{
	local station_list = GetAllInterCityStations();

	station_list.Valuate(AIStation.GetNearestTown);
	station_list.KeepValue(town_id);

	if(station_list.IsEmpty())
		return -1;

	local station = station_list.Begin();
	if(AITile.GetClosestTown(AIStation.GetLocation(station)) != town_id)
	{
		Helper.SetSign(GetLocation(town_id) + AIMap.GetTileIndex(2, 2), "  ERROR: station.GetNearestTown != tile.GetClosestTown");
		AILog.Error("@@@@@@@@@@@@ ERROR ERROR ERROR - station.GetNearestTown != tile.GetClosestTown for intercity station in town " + AITown.GetName(town_id))
	}

	return station_list.Begin();
}

//// END OF INTER-CITY

function GetGridStations(grid_list, town_stations)
{
	if(town_stations.IsEmpty())
		return town_stations;

	local ret_list = AIList();
	//local ret_list = AIStationList(AIStation.STATION_BUS_STOP);

	local grid_tile = grid_list.Begin();
	local i = 0;
	while(grid_list.HasNext())
	{
		town_stations.Valuate(AIStation.GetDistanceSquareToTile, grid_tile);
		town_stations.Sort(AIAbstractList.SORT_BY_VALUE, true); // lowest value first
		local closest_station = town_stations.Begin();
		
		local distance = town_stations.GetValue(town_stations.Begin());

		// Add stations that are close enough (only one station per grid_tile though)
		if(distance <= 2)
		{
			ret_list.AddItem(town_stations.Begin(), i++);
		}

		grid_tile = grid_list.Next();
	}

	ret_list.Sort(AIAbstractList.SORT_BY_VALUE, true); // lowest value first

	return ret_list;
}

function StationTileXValuator(station_id)
{
	return AIMap.GetTileX(AIStation.GetLocation(station_id));
}
function StationTileYValuator(station_id)
{
	return AIMap.GetTileY(AIStation.GetLocation(station_id));
}

function GetStationsRectangle(stations_list, grow_border_by = 0)
{
	//Helper.ClearAllSigns();

	local min_x, min_y, max_x, max_y;
	stations_list.Valuate(StationTileXValuator);
	stations_list.Sort(AIAbstractList.SORT_BY_VALUE, true); // lowest value first
	min_x = stations_list.GetValue(stations_list.Begin());
	stations_list.Valuate(StationTileXValuator);
	stations_list.Sort(AIAbstractList.SORT_BY_VALUE, false); // highest value first
	max_x = stations_list.GetValue(stations_list.Begin());

	stations_list.Valuate(StationTileYValuator);
	stations_list.Sort(AIAbstractList.SORT_BY_VALUE, true); // lowest value first
	min_y = stations_list.GetValue(stations_list.Begin());
	stations_list.Sort(AIAbstractList.SORT_BY_VALUE, false); // highest value first
	max_y = stations_list.GetValue(stations_list.Begin());
	
	local rect_from = AIMap.GetTileIndex(min_x - grow_border_by, min_y - grow_border_by);
	local rect_to = AIMap.GetTileIndex(max_x + grow_border_by, max_y + grow_border_by);

	//AISign.BuildSign(rect_from, "rect from");
	//AISign.BuildSign(rect_to, "rect to");

	local tiles = AITileList();
	tiles.AddRectangle(rect_to, rect_from);

	return tiles;
}

function BuildDepotAtTileScoreValuator(tile_id)
{
	local score = 0;

	local road_neighbour_count = AIRoad.GetNeighbourRoadCount(tile_id);

	if(AIRoad.IsRoadTile(tile_id))
	{
		// give score if tile has road, but only one neighbor has that, as that means it is a road stump
		if(road_neighbour_count == 1)
			score += 1000;
	}
	else
	{
		// better build it along an existing road, than hijacking a road stump
		if(road_neighbour_count > 0)
			score += 2000;
	}

	if(AITile.GetSlope(tile_id) == 0)
		score += 500;

	if(AITile.IsBuildable(tile_id))
	{
		score += 100;
	}
	else
	{
		local testmode = AITestMode();
		local acount = AIAccounting();

		AITile.DemolishTile(tile_id);
		
		score -= acount.GetCosts();
	}

	return score;
}

function GetDepotPlacementArea(town_id)
{
	local town_tiles = AITileList();
	local stations = Helper.GetBusOnlyStationsInTown(town_id);
	town_tiles = GetStationsRectangle(stations, 2);

	return town_tiles;
}

function FindDepot(tiles_to_search)
{
	local list_copy = AITileList();
	list_copy.AddList(tiles_to_search);

	// look for depot
	
	// To not modify the list, we sort the list with tiles with depot first
	list_copy.Valuate(AIRoad.IsRoadDepotTile);
	list_copy.KeepValue(1);
	//list_copy.Sort(AIAbstractList.SORT_BY_VALUE, false) // highest value first (1)

	// Only find own depot!
	list_copy.Valuate(AITile.GetOwner);
	list_copy.KeepValue(AICompany.ResolveCompanyID(AICompany.COMPANY_SELF));

	// and then check if first tile has a depot
	if(!list_copy.IsEmpty())
	{
		// and if so, return first tile.
		return list_copy.Begin();
	}
	
	// else return invalid tile
	//AILog.Info("returning tile -1, -1");
	return AIMap.GetTileIndex(-1, -1);
}

function GetRoadStationCatchedCargo(tile_id)
{
	return AITile.GetCargoAcceptance(tile_id, Helper.GetPAXCargo(), 1, 1, 3);
}

function Builder::ConstructTownStations(town_id)
{
	//Helper.ClearAllSigns();
	AILog.Info("building in " + AITown.GetName(town_id));
	//AISign.BuildSign(AITown.GetLocation(town_id), "Build stations");


	//// First build grid stations in town ////

	local tile_list = MakeTownGrid(town_id, GetBuildGridSize(town_id));
	local old_stations_list = Helper.GetBusOnlyStationsInTown(town_id);
	old_stations_list = GetGridStations(tile_list, old_stations_list);

	local new_stations_list = AIList();

	local tile = tile_list.Begin();
	local i = 0;
	while(tile_list.HasNext())
	{
		//AISign.BuildSign(tile, " grid " + i);

		local cargo = GetRoadStationCatchedCargo(tile);

		// If an old station is on grid-tile or one of the neighbours, then don't build a station
		old_stations_list.Valuate(AIStation.GetDistanceSquareToTile, tile);
		old_stations_list.Sort(AIAbstractList.SORT_BY_VALUE, true); // lowest value first
		if(tile_list.Count() > 1 &&  // don't construct anything if only one grid point has enough cargo
					(old_stations_list.IsEmpty() || old_stations_list.GetValue(old_stations_list.Begin()) > 2))
		{
			if(cargo >= 30) 
			{
				local ret = BuildInCityStationCloseTo(tile, town_id);
				if(AIMap.IsValidTile(ret))
				{
					new_stations_list.AddItem(ret, 0);
				}
			}
		}
		else if(tile_list.Count() <= 1 || (!old_stations_list.IsEmpty() && cargo < 15))
		{
			local station_tile = AIStation.GetLocation(old_stations_list.Begin());
			AIRoad.RemoveRoadStation(station_tile);
		}

		i++;
		tile = tile_list.Next();
	}

	// If failed to build new stations to reach 2 stations in town
	local total_stations_list = Helper.GetBusOnlyStationsInTown(town_id);
	if(total_stations_list.Count() < 2)
	{
		// don't leave any stations left
		//if(total_stations_list.Count() > 0)
		//	AITile.DemolishTile(AIStation.GetLocation(total_stations_list.Begin()));

		return false;
	}


	//// Second build depot ////

	// Check if the town has a depot
	local depot_placement_tiles = GetDepotPlacementArea(town_id);
	local depot = FindDepot(depot_placement_tiles);

	if(AIMap.IsValidTile(depot))
	{
		local depot_tile = depot_placement_tiles.Begin();
		AILog.Info("depot exist on tile " + depot_tile);
	}
	else
	{
		AILog.Info("building new depot");
	
		// get the boundary box of the stations in town

		// locate tile within depot_placement_tiles to place depot
		depot_placement_tiles.Valuate(AIRoad.GetNeighbourRoadCount);
		//depot_placement_tiles.KeepAboveValue(0);
		depot_placement_tiles.Valuate(AITile.IsWaterTile);
		//depot_placement_tiles.KeepValue(0);
		depot_placement_tiles.Valuate(AITile.IsStationTile);
		//depot_placement_tiles.KeepValue(0);
		//depot_placement_tiles.Valuate(AITile.IsSteepSlope);
		//depot_placement_tiles.KeepValue(0);
		depot_placement_tiles.Valuate(AITile.GetOwner);
		depot_placement_tiles.RemoveValue(AICompany.ResolveCompanyID(AICompany.COMPANY_SELF));

		MyValuate(depot_placement_tiles, BuildDepotAtTileScoreValuator);
		//depot_placement_tiles.Valuate(BuildDepotAtTileScoreValuator);
		depot_placement_tiles.Sort(AIAbstractList.SORT_BY_VALUE, false); // highest value first
		
		local depot_tile = depot_placement_tiles.Begin();
		while(depot_placement_tiles.HasNext())
		{
			//AILog.Info("Trying to build depot at tile " + depot_tile);
			//AISign.BuildSign(depot_tile, "build depot");

			local depot = true;
			if(BuildRoadStation(depot_tile, depot))
				break;			

			depot_tile = depot_placement_tiles.Next();
		}

		if(!AIRoad.IsRoadDepotTile(depot_tile))
			AILog.Error("Failed to build depot!");

		// depot built likely is equal to no buses yet in town.
		BuyBusesForTown(town_id)
	}

	//// Third decide if inter city station should be built ////

	local inter_city_station = FindInterCityStation(town_id);
	/*if(AIStation.IsValidStation(inter_city_station))
	{
		Helper.SetSign(AITown.GetLocation(town_id), "Has INTER CITY");
	}*/
	if(!AIStation.IsValidStation(inter_city_station) && AITown.GetPopulation(town_id) > 700)
	{
		Helper.SetSign(AITown.GetLocation(town_id), "trying to build IC");
		AILog.Info("Trying to build INTER-CITY station");

		local best_airport_available = GetBestAirportAvailable(8, 8);
		local ap_w = AIAirport.GetAirportWidth(best_airport_available);
		local ap_h = AIAirport.GetAirportHeight(best_airport_available);

		AILog.Info("best available airport = " + best_airport_available);

		ap_h += 2; // for bus stops

		
		local inter_city_placement = FindInterCityStationPlacement(town_id, ap_w, ap_h);
		if(AIMap.IsValidTile(inter_city_placement))
		{
			local ap_x = AIMap.GetTileX(inter_city_placement);
			local ap_y = AIMap.GetTileY(inter_city_placement);

			//Helper.ClearAllSigns();
			//AISign.BuildSign(inter_city_placement, "INTER-CITY");
			
			// Build airport
			local airport_status = AIAirport.BuildAirport(inter_city_placement, best_airport_available, AIStation.STATION_NEW);

			if(!AIAirport.IsAirportTile(inter_city_placement))
			{
				AILog.Info("Build airport error msg: " + AIError.GetLastErrorString());
			}
			else
			{
				// set station name
				local ic_station_id = AIStation.GetStationID(inter_city_placement);
				AIStation.SetName(ic_station_id, AITown.GetName(town_id) + " IC Airport");
	
				// Build bus stations
				local bus_y = ap_y + ap_h - 2;
				for(local bus_x = ap_x; bus_x < ap_x + ap_w; bus_x++)
				{
					local bus_tile = AIMap.GetTileIndex(bus_x, bus_y);
					local bus_front_tile = AIMap.GetTileIndex(bus_x, bus_y + 1);

					AIRoad.BuildRoad(bus_tile, bus_front_tile)

					AIRoad.BuildRoadStation(bus_tile, bus_front_tile, AIRoad.ROADVEHTYPE_BUS, AIStation.STATION_JOIN_ADJACENT);
				}

				// Build road infront of bus stations that connect all stations
				local road_from = AIMap.GetTileIndex(ap_x, bus_y + 1);
				local road_to = AIMap.GetTileIndex(ap_x + ap_w - 1, bus_y + 1);
				AIRoad.BuildRoad(road_from, road_to);

				// Connect inter-city station with town
				RoadBuilder.ConnectTiles(road_from, AITown.GetLocation(town_id), 5);

				// Change buses to feeder-mode
				local grid_stations_list = Helper.GetBusOnlyStationsInTown(town_id); // re-generate grid stations list
				grid_stations_list = GetGridStations(tile_list, grid_stations_list);

				local inter_city_station = FindInterCityStation(town_id);
				local vehicles =  GetAllVehiclesOfStationList(total_stations_list)

				if(AIStation.IsValidStation(inter_city_station) && vehicles.Count() > 0)
					UpdateIntraCityOrders_FeederMode(vehicles.Begin(), town_id, grid_stations_list, inter_city_station)
			}
		}
	}

	// No depot probably means there is no buses, but don't assume that, to be on the safe side.
	UpdateIntraCityOrdersForTown(town_id);

	return true;
}

function GetAllVehiclesOfStationList(station_list)
{
	local all_vehicles_list = AIList();
	local station = station_list.Begin();
	while(station_list.HasNext())
	{
		local vehicles = AIVehicleList_Station(station);
		all_vehicles_list.AddList(vehicles);

		station = station_list.Next();
	}

	return all_vehicles_list;
}
function GetAllRoadVehiclesOfStationList(station_list)
{
	local all_vehicles_list = GetAllVehiclesOfStationList(station_list);
	all_vehicles_list.Valuate(AIVehicle.GetVehicleType);
	all_vehicles_list.KeepValue(AIVehicle.VT_ROAD);

	return all_vehicles_list;
}

function GetAllAirVehiclesOfStationList(station_list)
{
	local all_vehicles_list = GetAllVehiclesOfStationList(station_list);
	all_vehicles_list.Valuate(AIVehicle.GetVehicleType);
	all_vehicles_list.KeepValue(AIVehicle.VT_AIR);

	return all_vehicles_list;
}

function GetBusEngineScoreValuator(engine_id, needed_capacity)
{
	local score = 0;

	// if can afford engine
	if(AIEngine.GetPrice(engine_id) < AICompany.GetBankBalance(AICompany.COMPANY_SELF))
		score += 1000;

	// If we need 20 capacity, then only those 20 should share the costs.
	local tot_pax_cap = AIEngine.GetCapacity(engine_id);
	local count_pax = Helper.Min(tot_pax_cap, needed_capacity)

	local price = AIEngine.GetPrice(engine_id);
	local running_cost = AIEngine.GetRunningCost(engine_id);

	local price_per_pax = price / count_pax;
	local running_cost_per_pax = running_cost / count_pax;


	score -= running_cost_per_pax * 3;

	score -= price_per_pax

	// some hacks to valuate capacity and speed more if needed capacity is really high
	local pax_factor = 2;
	local speed_factor = 2;
	if(needed_capacity > 200)
		pax_factor = 4;
	if(needed_capacity > 800)
	{
		speed_factor = 4;
		pax_factor = 8;
	}

	score += tot_pax_cap * pax_factor; // while only the needed pax cap shares the cost, a bigger bus get a score bonus even if the extra space isn't needed for now.

	score += AIEngine.GetMaxSpeed(engine_id) * speed_factor;

	score += AIBase.RandRange(100);

	if(AIEngine.GetReliability(engine_id) > 50)
		score += AIEngine.GetReliability(engine_id) * 2;
	
	return score;
}

function Builder::GetEngine(needed_capacity, vehicle_type)
{
	local engine_list = AIEngineList(vehicle_type);

	// Only keep non-articulated
	if(vehicle_type = AIVehicle.VT_ROAD)
	{
		engine_list.Valuate(AIEngine.IsArticulated)
		engine_list.KeepValue(0);
	}
	
	//engine_list.Valuate(AIEngine.GetCargoType);
	//engine_list.KeepValue(Helper.GetPAXCargo());
	engine_list.Valuate(AIEngine.CanRefitCargo, Helper.GetPAXCargo());
	engine_list.KeepValue(1);

	/*for(local i = engine_list.Begin(); engine_list.HasNext(); i = engine_list.Next())
	{
		AILog.Info("Engine name: " + AIEngine.GetName(i));
	}*/

	MyValuate(engine_list, GetBusEngineScoreValuator, needed_capacity);
	engine_list.Sort(AIAbstractList.SORT_BY_VALUE, false); // highest first
	engine_list.KeepTop(1);
	return engine_list.Begin();
}

// A function to clear all orders of a vehicle, that works so that if shared orders
// is used, each of the shared order is removed, but the sharing of the orders is
// not affected - only clearing of them.
function ClearOrders(vehicle_id)
{
	local num_orders = AIOrder.GetOrderCount(vehicle_id);
	for(local i = 0; i < num_orders; i++)
	{
		AIOrder.RemoveOrder(vehicle_id, 0);
	}
	if(AIOrder.GetOrderCount(vehicle_id) != 0)
		AILog.Error("Not all errors cleared");
}

function Builder::GetStationListFromOrders(vehicle_id)
{
	local station_list = AIList();

	local num_orders = AIOrder.GetOrderCount(vehicle_id);
	for(local i = 0; i < num_orders; i++)
	{
		local tile_id = AIOrder.GetOrderDestination(vehicle_id, i);
		local station_id = AIStation.GetStationID(tile_id);
		if(AIStation.IsValidStation(station_id))
		{
			station_list.AddItem(station_id, 0);
		}
	}

	return station_list;
}

function Builder::HasStationInOrders(vehicle_id, station_id)
{
	local station_list = GetStationListFromOrders(vehicle_id);
	return station_list.HasItem(station_id);
}

// Function that makes sure that the vehicle has the stations in the stations_list
// as orders. And when adding new orders it uses non-stop flag
function Builder::MakeSureVehicleHasTheseStations(vehicle_id, station_list, new_order_flags)
{
	if(!AIVehicle.IsValidVehicle(vehicle_id))
	{
		AILog.Warning("MakeSureVehicleHasTheseStations: Vehicle is INVALID");
	}
	if(station_list.IsEmpty())
	{
		AILog.Warning("MakeSureVehicleHasTheseStations: Station list is empty");
	}

	local missing_stations = AIList();
	for(local station_id = station_list.Begin(); station_list.HasNext(); station_id = station_list.Next())
	{
		if(!HasStationInOrders(vehicle_id, station_id))
		{
			missing_stations.AddItem(station_id, 0);
		}
	}
	
	for(local station_id = missing_stations.Begin(); missing_stations.HasNext(); station_id = missing_stations.Next())
	{
		AIOrder.AppendOrder(vehicle_id, AIStation.GetLocation(station_id), new_order_flags);
	}
}

// Intra (in city) transport
function Builder::UpdateIntraCityOrders_LocalMode(vehicle_id, town_id, town_stations)
{
	// remove inter city station if it is in orders
	local inter_city_station = FindInterCityStation(town_id);
	if(AIStation.IsValidStation(inter_city_station) && HasStationInOrders(vehicle_id, inter_city_station))
	{
		ClearOrders(vehicle_id);
		/*for(local order_id = 0; order_id < AIOrder.GetOrderCount(vehicle_id); order_id++)
		{
			if(AIOrder.GetOrderDestination(vehicle_id, order_id) == inter_city_station)
				AIOrder.RemoveOrder(order_id);
		}*/
	}

	MakeSureVehicleHasTheseStations(vehicle_id, town_stations, AIOrder.AIOF_NON_STOP_INTERMEDIATE);
}

function Builder::UpdateIntraCityOrders_FeederMode(vehicle_id, town_id, bus_only_stations, airport_station)
{
	AILog.Warning("UpdateIntraCityOrders_FeederMode, vehicle: " + AIVehicle.GetName(vehicle_id));

	local order_list = OrderList();

	for(local station_id = bus_only_stations.Begin(); bus_only_stations.HasNext(); station_id = bus_only_stations.Next())
	{
		Helper.SetSign(AIStation.GetLocation(station_id), "add to feeder mode");
		order_list.AddStop(station_id, AIOrder.AIOF_NON_STOP_INTERMEDIATE | AIOrder.AIOF_NO_UNLOAD);
	}
	order_list.AddStop(airport_station, AIOrder.AIOF_NON_STOP_INTERMEDIATE | AIOrder.AIOF_NO_LOAD | AIOrder.AIOF_TRANSFER);

	order_list.ApplyToVehicle(vehicle_id);


/*	local has_airport_station_in_orders = HasStationInOrders(vehicle_id, airport_station);

	if(!has_airport_station_in_orders)
	{
		AILog.Info("clearing orders of vehicle ...");
		ClearOrders(vehicle_id);
		AILog.Info("clearing orders of vehicle - DONE");
	}

	MakeSureVehicleHasTheseStations(vehicle_id, bus_only_stations, AIOrder.AIOF_NON_STOP_INTERMEDIATE | AIOrder.AIOF_NO_UNLOAD);
	
	if(!has_airport_station_in_orders)
	{
		AILog.Info("Adding airport to orders");
		AIOrder.AppendOrder(vehicle_id, AIStation.GetLocation(airport_station), AIOrder.AIOF_NON_STOP_INTERMEDIATE | AIOrder.AIOF_NO_LOAD | AIOrder.AIOF_TRANSFER);
	}*/
}

/* Just use 
 * local inter_city_station = FindInterCityStation(town_id);
 * and then send inter_city_station as second parameter
 * 
 * If the function return true, then inter_city_station is valid too
 */
function Builder::ShouldHaveFeederMode(town_id, inter_city_station)
{
	// There need to be at least 2 airports and the inter_city_station need to be valid
	if(GetAllInterCityStations().Count() < 2 || !AIStation.IsValidStation(inter_city_station))
		return false;

	// but the inter city station in question should also be serviced by at least one airplane
	local station_list = AIList();
	station_list.AddItem(inter_city_station, 0);
	local vehicle_list = GetAllAirVehiclesOfStationList(station_list);

	return !vehicle_list.IsEmpty();
}

function Builder::UpdateIntraCityOrdersForTown(town_id)
{
	local inter_city_station = FindInterCityStation(town_id);

	local stations_list = Helper.GetBusOnlyStationsInTown(town_id);
	local grid_list = MakeTownGrid(town_id, GetBuildGridSize(town_id));
	local grid_stations_list = GetGridStations(grid_list, stations_list);
	local station_vehicles_list = GetAllRoadVehiclesOfStationList(grid_stations_list);

	if(ShouldHaveFeederMode(town_id, inter_city_station))
	{
		// inter-city-mode -> feed the airport
		AILog.Warning("UpdateIntraCityOrdersForTown: Town " + AITown.GetName(town_id) + " is in INTER-CITY mode");

		if(!station_vehicles_list.IsEmpty())
		{
			AILog.Info("updating orders - feeder mode");
			UpdateIntraCityOrders_FeederMode(station_vehicles_list.Begin(), town_id, grid_stations_list, inter_city_station);
		}
	}
	else
	{
		if(!station_vehicles_list.IsEmpty())
		{
			AILog.Info("updating orders - local mode");
			UpdateIntraCityOrders_LocalMode(station_vehicles_list.Begin(), town_id, grid_stations_list);
		}
	}
}

// Inter city
function Builder::UpdateInterCityOrders(vehicle_id, inter_city_stations)
{
	MakeSureVehicleHasTheseStations(vehicle_id, inter_city_stations, AIOrder.AIOF_NONE);
}

function VehicleRunningCostPerCapacityValuator(vehicle_id, needed_capacity)
{
	local capacity = Helper.Min(needed_capacity, AIVehicle.GetCapacity(vehicle_id, Helper.GetPAXCargo())); // don't let capacity we don't need pay for the running cost
	local running_cost = AIVehicle.GetRunningCost(vehicle_id);

	if(capacity == 0)
		return running_cost * 1000;

	return running_cost / capacity;
}

function SendVehicleToDepotForSelling(vehicle_id)
{
	local start_tick = main_instance.GetTick();
	while(!AIVehicle.SendVehicleToDepot(vehicle_id))
	{
		if(start_tick + 100 < main_instance.GetTick())
		{
			AILog.Info("Failed to send vehicle " + AIVehicle.GetName(vehicle_id) + " to depot, for more than 100 ticks, aborting selling of this vehicle.");
			return false;
		}

		AILog.Info("Failed to send vehicle " + AIVehicle.GetName(vehicle_id) + " to depot, retrying soon .. | start_tick = " + start_tick + ", curr tick = " + main_instance.GetTick());
		main_instance.Sleep(5);
	}

	AIOrder.UnshareOrders(vehicle_id);
	ClearOrders(vehicle_id);
}

function BuyVehicles(existing_capacity, target_capacity, engine, depot, existing_vehicles_list, order_data)
{
	local needed_capacity = target_capacity - existing_capacity;
	local engine_capacity = AIEngine.GetCapacity(engine);
	local vehicle_type = AIEngine.GetVehicleType(engine);


	if(AICompany.GetBankBalance(AICompany.COMPANY_SELF) < AIEngine.GetPrice(engine) + 10000)
		return;

	// Buy one engine
	local vehicle = AIVehicle.BuildVehicle(depot, engine);
	if(!AIVehicle.IsValidVehicle(vehicle))
	{
		AILog.Error("failed to buy vehicle");
		return false;
	}
	AIVehicle.RefitVehicle(vehicle, Helper.GetPAXCargo()); // since GetEngine can return engines that by default don't carry PAX (but can be refitted to PAX), we need to refit bough vehicles to PAX.
	AIVehicle.SetName(vehicle, "Re: " + AIError.GetLastErrorString());
	

	AIVehicle.StartStopVehicle(vehicle);

	// Make orders
	if(!existing_vehicles_list.IsEmpty())
	{
		AILog.Info("Share orders. New vehicle " + AIVehicle.GetName(vehicle) + " shares from " + AIVehicle.GetName(existing_vehicles_list.Begin()));
		AIOrder.ShareOrders(vehicle, existing_vehicles_list.Begin());
	}
	else
	{
		// different vehicle types (buses vs airplanes) has different oder functions
		if(vehicle_type == AIVehicle.VT_ROAD)
		{
			AILog.Info("First vehicle (road) so make new orders. feeder_mode = " + order_data.feeder_mode? "true":"false");
			if(order_data.feeder_mode)
				UpdateIntraCityOrders_FeederMode(vehicle, order_data.town_id, order_data.grid_stations_list, order_data.inter_city_station);
			else
				UpdateIntraCityOrders_LocalMode(vehicle, order_data.town_id, order_data.grid_stations_list);
		}
		else if(vehicle_type == AIVehicle.VT_AIR)
		{
			AILog.Info("First vehicle (air) so make new orders.");

			MakeSureVehicleHasTheseStations(vehicle, order_data.station_list, AIOrder.AIOF_NONE);
		}
	}

	// Clone the first built vehicle until needed capacity is reached or cloning fails
	needed_capacity -= engine_capacity;
	while(needed_capacity > 0 && 
			AICompany.GetBankBalance(AICompany.COMPANY_SELF) < AIEngine.GetPrice(engine) * 2) // a little buffer of money is good to have
	{
		AILog.Info("Clone vehicle");
		local share_orders = true;
		local clone_vehicle = AIVehicle.CloneVehicle(depot, vehicle, share_orders);

		if(!AIVehicle.IsValidVehicle(clone_vehicle))
		{
			AILog.Error("failed to clone vehicle");
			return false;
		}

		AIVehicle.StartStopVehicle(clone_vehicle);
		needed_capacity -= engine_capacity;

		existing_vehicles_list.AddItem(clone_vehicle, 0); // add vehicle to list, in case caller have use for it
	}

	return true;
}

function SellVehicles(existing_capacity, target_capacity, existing_vehicles_list)
{
	local needed_capacity = target_capacity - existing_capacity;

	local keep_vehicle_list = AIList();
	local sell_vehicle_list = AIList();
	sell_vehicle_list.AddList(existing_vehicles_list);
	local kept_capacity = 0;
	while(kept_capacity < target_capacity)
	{
		// sort vehicles after their running cost / pax. But as when buying vehicles, if we only need 20, and the bus has 30 seats, only
		// let the needed 20 seats share the cost.
		sell_vehicle_list.Valuate(VehicleRunningCostPerCapacityValuator, target_capacity - kept_capacity);

		sell_vehicle_list.Sort(AIAbstractList.SORT_BY_VALUE, true); // lowest value first
		local keep_vehicle = sell_vehicle_list.Begin();
		sell_vehicle_list.RemoveTop(1);
		
		keep_vehicle_list.AddItem(keep_vehicle, 0);
		kept_capacity += AIVehicle.GetCapacity(keep_vehicle, Helper.GetPAXCargo());
	}

	if(sell_vehicle_list.IsEmpty())
		AILog.Warning("Warning: sell vehicles list is Empty!");

	// now sell the vehicles in sell_vehicle_list
	for(local vehicle_id = sell_vehicle_list.Begin(); sell_vehicle_list.HasNext(); vehicle_id = sell_vehicle_list.Next())
	{
		SendVehicleToDepotForSelling(vehicle_id);
	}
}

function Builder::BuyBusesForTown(town_id)
{
	local stations_list = Helper.GetBusOnlyStationsInTown(town_id);
	local grid_list = MakeTownGrid(town_id, GetBuildGridSize(town_id));
	local grid_stations_list = GetGridStations(grid_list, stations_list);

	AILog.Info("Buying new buses for town " + AITown.GetName(town_id) + " that has " + stations_list.Count() + " stations");

	if(grid_stations_list.Count() < 1) 
	{
		return false;
	}

	// Check if feeder-mode or just transporting around passengers in the town
	// This affects some parts below
	local inter_city_station = FindInterCityStation(town_id);
	local feeder_mode = ShouldHaveFeederMode(town_id, inter_city_station);

	local station_vehicles_list = GetAllRoadVehiclesOfStationList(grid_stations_list);
	local town_pop = AITown.GetPopulation(town_id);

	local num_vehicles_per_station = station_vehicles_list.Count() / grid_stations_list.Count();

	// Decide how many buses are needed
	station_vehicles_list.Valuate(AIVehicle.GetCapacity, Helper.GetPAXCargo());
	local existing_capacity = Helper.ListValueSum(station_vehicles_list);
	local target_capacity = existing_capacity;

	if(grid_stations_list.Count() < 2 || (feeder_mode && grid_stations_list.Count() < 1))
	{
		target_capacity = 0;
	}
	else if(station_vehicles_list.IsEmpty())
	{
		// if there are no buses, base it on town population
		target_capacity = town_pop / 20;
	}
	else
	{
		// check how well the service work now and adjust the capacity if needed
		grid_stations_list.Valuate(AIStation.GetCargoWaiting, Helper.GetPAXCargo());
		grid_stations_list.Sort(AIAbstractList.SORT_BY_VALUE, false); // highest first
		local max_waiting = grid_stations_list.GetValue(grid_stations_list.Begin());
		local tot_waiting = Helper.ListValueSum(grid_stations_list);
		local mean_waiting = tot_waiting / grid_stations_list.Count();

		if(max_waiting > 150 || mean_waiting > 60)
		{
			target_capacity = existing_capacity + 50 + Helper.Max(0, mean_waiting - 125);	
		}
		else if(mean_waiting < 20)
		{
			target_capacity = Helper.Max(existing_capacity - 50, existing_capacity / 2);
		}
	}

	// Don't buy toooo many buses if the stations can't handle more.
	if(target_capacity != 0 && num_vehicles_per_station > conf_max_buy_num_buses_per_bus_station)
	{
		target_capacity = existing_capacity;
	}

	// In feeder mode,
	if(target_capacity != 0 && feeder_mode)
	{
		local ic_station_waiting = AIStation.GetCargoWaiting(inter_city_station, Helper.GetPAXCargo());
		local use_break_points = Helper.HasSign(AITown.GetName(town_id) + "_debug");

		AILog.Info("Waiting PAX at IC station = " + ic_station_waiting);
		AILog.Info("target_capacity = " + target_capacity);
		if(use_break_points)
			Helper.BreakPoint(AIStation.GetLocation(inter_city_station) + 1);

		// don't buy more buses when there is more than 200 waiting -> instead spend the money on planes
		if(ic_station_waiting > 200)
			target_capacity = existing_capacity;

		AILog.Info("target_capacity = " + target_capacity);
		if(use_break_points)
			Helper.BreakPoint(AIStation.GetLocation(inter_city_station) + 1);
		
		// and reduce the number of buses if there is more than 700 waiting
		if(ic_station_waiting > 400)
			target_capacity = Helper.Max(existing_capacity - 50, existing_capacity / 2);

		AILog.Info("target_capacity = " + target_capacity);
		if(use_break_points)
			Helper.BreakPoint(AIStation.GetLocation(inter_city_station) + 1);
	}

	local needed_capacity = target_capacity - existing_capacity;
	if(needed_capacity > 0) 
	{
		AILog.Info("Need more buses because town has capacity: " + existing_capacity + " but want to have capacity: " + target_capacity);

		// bus engine
		local bus_engine = GetEngine(needed_capacity, AIVehicle.VT_ROAD);

		// Find a depot
		local depot_placement_tiles = GetDepotPlacementArea(town_id);
		local depot = FindDepot(depot_placement_tiles);

		if(!AIMap.IsValidTile(depot))
		{
			AILog.Error("can't find depot in town " + AITown.GetName(town_id));
			return false;
		}

		// order data
		local orders_data = { town_id = town_id, feeder_mode = feeder_mode, grid_stations_list = grid_stations_list, inter_city_station = inter_city_station };

		return BuyVehicles(existing_capacity, target_capacity, bus_engine, depot, station_vehicles_list, orders_data);
	}
	else if(needed_capacity < 0)
	{
		AILog.Info("Want to reduce capacity from: " + existing_capacity + " to target capacity: " + target_capacity);

		SellVehicles(existing_capacity, target_capacity, station_vehicles_list);
	}
	else
	{
		AILog.Info("has enough capacity: " + existing_capacity + " for target capacity: " + target_capacity);
	}

	return true;
}

function GetHangarTile(station_id)
{
	local airport_tiles = AITileList_StationType(station_id, AIStation.STATION_AIRPORT);
	if(airport_tiles.IsEmpty())
		return -1;

	local hangar_tile = AIAirport.GetHangarOfAirport(airport_tiles.Begin());
	return hangar_tile;
}

function Builder::ManageInterCityVehiclesOfInterCityConnection(station_list)
{
	AILog.Info("Builder::ManageInterCityVehiclesOfInterCityConnection");

	if(station_list.IsEmpty())
	{
		AILog.Info("Builder::ManageInterCityVehiclesOfInterCityConnection: Connection is empty");
		return false;
	}

	for(local i = station_list.Begin(); station_list.HasNext(); i = station_list.Next())
	{
		AILog.Info("Station: " + AIStation.GetName(i));
	}

	// Get vehicle list
	local all_vehicles_list = GetAllAirVehiclesOfStationList(station_list);
	AILog.Info("Builder::ManageInterCityVehiclesOfInterCityConnection: has " + all_vehicles_list.Count() + " vehicles and " + station_list.Count() + " stations");

	// Get age of youngest plane
	all_vehicles_list.Valuate(AIVehicle.GetAge);
	all_vehicles_list.Sort(AIAbstractList.SORT_BY_VALUE, true); // youngest first
	local youngest_airplane = all_vehicles_list.GetValue(all_vehicles_list.Begin());

	if(all_vehicles_list.IsEmpty() || youngest_airplane > 30)
	{
		// Get existing capacity
		all_vehicles_list.Valuate(AIVehicle.GetCapacity, Helper.GetPAXCargo());
		local existing_capacity = Helper.ListValueSum(all_vehicles_list);
		local target_capacity = existing_capacity;

		// Get number of waiting passengers
		station_list.Valuate(AIStation.GetCargoWaiting, Helper.GetPAXCargo());
		local sum_waiting = Helper.ListValueSum(station_list);
		local mean_waiting = sum_waiting / station_list.Count();
		station_list.Sort(AIAbstractList.SORT_BY_VALUE, false); // max waiting first
		local max_waiting_station_id = station_list.Begin();
		local max_waiting = station_list.GetValue(station_list.Begin());

		// Decide target capacity
		if(all_vehicles_list.IsEmpty())
		{
			// find an engine, where we pay for up to 100 seats, and set target capacity to the capacity of that engine
			local engine = GetEngine(100, AIVehicle.VT_AIR);
			target_capacity = AIEngine.GetCapacity(engine);
		}
		else if(mean_waiting > 100 || max_waiting > 200)
		{
			target_capacity = existing_capacity + 50 + Helper.Max(0, (mean_waiting - 100)*2 / 3);
		}
		else if(mean_waiting < 10)
		{
			target_capacity = Helper.Max(existing_capacity - 70, existing_capacity / 2);
		}

		// don't allow more than 4 planes per station <-- Hack, until better method to determine capacity problem is constructed
		local max_num_planes = station_list.Count() * 4;
		if(station_list.Count() == 2)
		{
			// but allow one extra per 40 tiles between the airports
			local a = station_list.Begin();
			local b = station_list.Next();
			
			local dist = AIMap.DistanceManhattan(AIStation.GetLocation(a), AIStation.GetLocation(b));
			max_num_planes += dist / 40;
		}
		if(all_vehicles_list.Count() >= max_num_planes && target_capacity >= existing_capacity)
		{
			target_capacity = existing_capacity
		}

		local needed_capacity = target_capacity - existing_capacity;
		if(needed_capacity > 0) 
		{
			AILog.Info("Need more airplanes, because connection has capacity: " + existing_capacity + " but want to have capacity: " + target_capacity);

			// airplane engine
			local airplane_engine = GetEngine(needed_capacity, AIVehicle.VT_AIR);

			// find a hangar
			local hangar_tile = GetHangarTile(station_list.Begin());
			if(!AIAirport.IsHangarTile(hangar_tile))
			{
				AILog.Warning("Failed to find hangar tile of station " + AIStation.GetName(station_list.Begin()));
				return false;
			}

			// order data
			local orders_data = { station_list = station_list };

			return BuyVehicles(existing_capacity, target_capacity, airplane_engine, hangar_tile, all_vehicles_list, orders_data);
		}
		else if(needed_capacity < 0)
		{
			AILog.Info("Want to reduce capacity from: " + existing_capacity + " to target capacity: " + target_capacity);

			SellVehicles(existing_capacity, target_capacity, all_vehicles_list);
		}
		else
		{
			AILog.Info("has enough capacity: " + existing_capacity + " for target capacity: " + target_capacity);
		}

	} // <- end of buy/sell airplanes

	all_vehicles_list = GetAllAirVehiclesOfStationList(station_list);
	if(!all_vehicles_list.IsEmpty())
	{
		UpdateInterCityOrders(all_vehicles_list.Begin(), station_list);
	}
}

function Builder::SellOldVehicles()
{
	AILog.Info("Looking for old vehicles to sell");
	local vehicle_list = AIVehicleList();

	AILog.Info("has " + vehicle_list.Count() + " vehicles");

	vehicle_list.Valuate(AIVehicle.GetAgeLeft);
	vehicle_list.KeepBelowValue(365*2);

	AILog.Info("has " + vehicle_list.Count() + " old vehicles");

	for(local vehicle_id = vehicle_list.Begin(); vehicle_list.HasNext(); vehicle_id = vehicle_list.Next()) 
	{
		AILog.Info("Selling old vehicle");
		SendVehicleToDepotForSelling(vehicle_id);
	}
}

function Builder::ScanForStopedVehiclesInDepots()
{
	local vehicle_list = AIVehicleList();
	vehicle_list.Valuate(AIVehicle.IsStoppedInDepot);
	vehicle_list.KeepValue(1);

	for(local vehicle_id = vehicle_list.Begin(); vehicle_list.HasNext(); vehicle_id = vehicle_list.Next()) 
	{
		// vehicles without orders in depot is a signal that they should be sold
		if(AIOrder.GetOrderCount(vehicle_id) == 0)
			AIVehicle.SellVehicle(vehicle_id);
	}
}
