/*
 * This file is part of PAXLink, which is an AI for OpenTTD
 * Copyright (C) 2008, 2009  Leif Linse
 *
 * PAXLink is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License
 *
 * PAXLink is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with PAXLink; If not, see <http://www.gnu.org/licenses/> or
 * write to the Free Software Foundation, Inc., 51 Franklin Street, 
 * Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

require("builder.nut");
require("helper.nut");
require("globals.nut");

class PAXLink extends AIController {
	builder = null;
	last_serviced_towns_check = null;
	last_depot_check = null;
	last_look_for_new_towns = null;
	last_inter_city_check = null;

	constructor()
	{
		main_instance = this; 
		this.builder = Builder();
		this.last_serviced_towns_check = 0;
		this.last_depot_check = 0;
		this.last_look_for_new_towns = 0;
		this.last_inter_city_check = 0;

	}

	function Start();
	function Stop()
	{
		Helper.ClearAllSigns();
	}
}

function PAXLink::SetCompanyName()
{
	AICompany.SetName("PAXLink")
	AICompany.SetPresidentName("PAX-King")

}

//// INTRA CITY TRANSPORT

function NumBusOnlyStationsInTownValuator(town_id)
{
	local stations = Helper.GetBusOnlyStationsInTown(town_id);
	
	return stations.Count();
}

function PAXLink::ManageIntraCityTransport()
{
	// look through cities that currently are serviced first
	local serviced_towns = AITownList();
	serviced_towns.Valuate(NumBusOnlyStationsInTownValuator);
	serviced_towns.KeepAboveValue(0);

	//AILog.Info("Has stations in " + serviced_towns.Count() + " towns.");

	// Look for vehicles in depot every 50 tick
	if(this.last_depot_check + 50 < this.GetTick())
	{
		this.last_depot_check = this.GetTick();
		this.builder.ScanForStopedVehiclesInDepots();
	}

	// Only check serviced town at a certain interval
	if(this.last_serviced_towns_check + 500 < this.GetTick())
	{
		this.last_serviced_towns_check = this.GetTick();

		AILog.Info("#### Checking the already serviced towns ####");

		this.builder.SellOldVehicles();

		// Start with town with biggest POP and walk down, for easier debugging
		serviced_towns.Valuate(AITown.GetPopulation);
		serviced_towns.Sort(AIAbstractList.SORT_BY_VALUE, false); // highest value first


		local bus_engine = this.builder.GetEngine(1000, AIVehicle.VT_ROAD);

		local town_id = serviced_towns.Begin();
		while(serviced_towns.HasNext())
		{
			this.builder.BuyBusesForTown(town_id);

			if(AICompany.GetBankBalance(AICompany.COMPANY_SELF) > AIEngine.GetPrice(bus_engine) * 2)
				this.builder.ConstructTownStations(town_id);

			town_id = serviced_towns.Next();
		}
	}

	// check if there are any non-serviced stations
	// ...
	if(this.last_look_for_new_towns + 500 < this.GetTick())
	{

		// Then look through cities that doesn't yet has any service
		// - But only if there is enough money or we don't have any serviced town yet and must try anyways.
		local bus_engine = this.builder.GetEngine(1000, AIVehicle.VT_ROAD);
		if(serviced_towns.Count() == 0 || AICompany.GetBankBalance(AICompany.COMPANY_SELF) > AIEngine.GetPrice(bus_engine) * 5) 
		{
			AILog.Info("#### Looking for new towns to service ####");

			local towns = AITownList();
			towns.Valuate(NumBusOnlyStationsInTownValuator);
			towns.KeepBelowValue(1);
			
			towns.Valuate(AITown.GetPopulation);
			towns.Sort(AIAbstractList.SORT_BY_VALUE, false); // highest value first
			//towns.KeepTop(1);

			if(serviced_towns.Count() != 0)
				towns.KeepAboveValue(450);
		
			for(local town = towns.Begin(); towns.HasNext(); town = towns.Next())
			{
				if(this.builder.ConstructTownStations(town))
				{
					this.last_look_for_new_towns = this.GetTick() - 400; // only wait 100 ticks (instead of 500) for next look for new town if found a town
					return;
				}

				AILog.Info("take next town");
			}
		}

		this.last_look_for_new_towns = this.GetTick();
	}
}

// Between cities
function PAXLink::ManageInterCityTransport()
{
	if(this.last_inter_city_check + 200 < this.GetTick())
	{
		this.last_inter_city_check = this.GetTick();

		local bus_engine = this.builder.GetEngine(1000, AIVehicle.VT_ROAD);

		local serviced_towns = AITownList();
		serviced_towns.Valuate(NumBusOnlyStationsInTownValuator);
		serviced_towns.KeepAboveValue(0);

		if(serviced_towns.Count() >= 2)
		{
			AILog.Info("#### Managing inter-city transport ####");

			local towns_with_inter_city = AITownList();
			towns_with_inter_city.Valuate(AITown.GetPopulation);
			towns_with_inter_city.Sort(AIAbstractList.SORT_BY_VALUE, false); // highest pop first

			//this.builder.FindInterCityStation(towns_with_inter_city.Begin());
			Helper.MyClassValuate(towns_with_inter_city, this.builder.FindInterCityStation, this.builder);

			towns_with_inter_city.RemoveValue(-1);

			local towns_to_connect = AITownList();
			towns_to_connect.Valuate(NumBusOnlyStationsInTownValuator);
			towns_to_connect.KeepAboveValue(1);
			towns_to_connect.Valuate(AITown.GetPopulation);
			towns_to_connect.KeepAboveValue(700);
			
			if(towns_with_inter_city.Count() > 1)
			{
				local stations_need_connection = AIList();

				// First go through existing inter-city stations and check if any of them has no service
				for(local town_id = towns_with_inter_city.Begin(); towns_with_inter_city.HasNext(); town_id = towns_with_inter_city.Next())
				{
					local inter_city_station_id = towns_with_inter_city.GetValue(town_id);
					local vehicles = AIVehicleList_Station(inter_city_station_id);
					vehicles.Valuate(AIVehicle.GetVehicleType);
					vehicles.KeepValue(AIVehicle.VT_AIR);

					if(vehicles.IsEmpty())
					{
						stations_need_connection.AddItem(inter_city_station_id, 0);
					}
					else
					{
						local vehicle_id = vehicles.Begin();
						local connection_stations = this.builder.GetStationListFromOrders(vehicles.Begin());
						AILog.Info("Manage vehicles of existing connection");
						this.builder.ManageInterCityVehiclesOfInterCityConnection(connection_stations);

						towns_with_inter_city.RemoveList(connection_stations); // TODO: Check if this is safe
					}
				}

				// as towns_with_inter_city was sorted with highest pop first
				// then stations_need_connection should be sorted this way too
				while(stations_need_connection.Count() > 1)
				{
					local station_1 = stations_need_connection.Begin();
					stations_need_connection.RemoveTop(1);
					
					stations_need_connection.Valuate(AIMap.DistanceSquare, station_1);
					stations_need_connection.Sort(AIAbstractList.SORT_BY_VALUE, false); // as far a way as possible

					local station_2 = stations_need_connection.Begin();
					stations_need_connection.RemoveTop(1);

					local new_connection = AIList();
					new_connection.AddItem(station_1, 0);
					new_connection.AddItem(station_2, 0);
					AILog.Info("Manage vehicles of new connection");
					this.builder.ManageInterCityVehiclesOfInterCityConnection(new_connection);

					// sort with highest pop first for next round
					if(stations_need_connection.Count() > 1)
					{
						stations_need_connection.Valuate(AITown.GetPopulation);
						stations_need_connection.Sort(AIAbstractList.SORT_BY_VALUE, false); // highest pop first
					}
				}
			}
			else
			{
				AILog.Info("Not enough towns has inter-city stations");
			}
		}
	}
}

function PAXLink::Start()
{
	this.Sleep(1);
	this.SetCompanyName();

	while (true) {
		AICompany.SetLoanAmount(AICompany.GetMaxLoanAmount());

		ManageIntraCityTransport();
		this.Sleep(5);
		ManageInterCityTransport();
		this.Sleep(5);

	}
}

